/* pxy.c: open proxy checker, proxy protocol routines.
 * Michael Tokarev  <mjt@corpit.ru>.
 * This code may be freely used and distributed according to
 * the terms of General Public License (GPL) version 2 or later.
 */

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "event.h"
#include "pxy.h"

int pxytimeout = 30000; /* 30 sec */
unsigned char pxybuf[8193];

static void timedout(int UNUSED fd, ev_t UNUSED e, void *ctx) {
  errno = ETIMEDOUT;
  ev_iowakeup((int)ctx, EVENT_ERROR);
}

int pxyrenew(int fd, pxyconn_t *c, int tmo) {
  if (ev_settimer(tmo ? tmo : pxytimeout, timedout, (void*)fd) == 0)
    return 0;
  else {
    pxyaction(fd, c, -1);
    return -1;
  }
}

int pxycerr(int fd, ev_t e, pxyconn_t *c) {
  if (e & EVENT_RDWR)
    return 0;
  /*
  if (errno != ETIMEDOUT)
    pxyinfo(c, 1, "error: %s", strerror(errno));
  */
  pxyaction(fd, c, -1);
  return -1;
}

pxyconn_t *pxynew() {
  pxyconn_t *c = (pxyconn_t *)malloc(sizeof(pxyconn_t));
  if (c)
    memset(c, 0, sizeof(*c));
  return c;
}

void pxyfree(int fd, pxyconn_t *c) {
  if (fd >= 0) {
    ev_settimer(-1, timedout, (void*)fd);
    ev_reqio(fd, 0, NULL, NULL);
  }
  if (c->buf) {
    free(c->buf);
    c->buf = NULL;
  }
  free(c);
}

static int
pxyev_reqio(int fd, ev_t e, void(*fn)(int,ev_t,pxyconn_t*), pxyconn_t *c) {
  if (ev_reqio(fd, e, (void(*)(int,ev_t,void*))fn, c) == 0)
    return 0;
  else {
    pxyaction(fd, c, -1);
    return -1;
  }
}

int pxywrite(int fd, pxyconn_t *c, const char *buf, int len, int level) {
  int r;
  if ((r = write(fd, buf, len)) <= 0) {
    pxyaction(fd, c, r < 0 ? -1 : 1);
    return -1;
  }
  else {
    pxyvio(c, level, 1, buf, len);
    return 0;
  }
}

int pxystart(int fd, pxyconn_t *c) {
  return
    pxyev_reqio(fd, EVENT_WRITE, c->proto->handler, c) == 0 &&
    pxyrenew(fd, c, 0) == 0 ?
    0 : -1;
}

int pxyread(int fd, pxyconn_t *c, char *buf, int l, int level) {
  errno = 0;
  l = read(fd, buf, l);
  if (l < 0) {
    pxyaction(fd, c, -1);
    return -1;
  }
  if (l > 0) {
    pxyvio(c, level, 0, buf, l);
    c->nread += l;
  }
  buf[l] = '\0';
  return l;
}

int pxyreadnext(int fd, pxyconn_t *c, int level) {
  int l, r;
  if (c->buf && (l = c->buf->len) != 0)
    memcpy(pxybuf, c->buf->buf, l);
  else
    l = 0;
  r = pxyread(fd, c, pxybuf + l, sizeof(pxybuf) - l - 1, level);
  return r <= 0 ? r : r + l;
}

int pxysave(int fd, pxyconn_t *c, char *buf, unsigned len, unsigned max) {
  if (c->nread > 2048) {
    pxyinfo(c, 2, "too much input");
    pxyaction(fd, c, 2);
    return -1;
  }
  if (buf && len > 0) {
    if (!c->buf && !(c->buf = (pxybuf_t*)malloc(sizeof(pxybuf_t)))) {
      errno = ENOMEM;
      pxyaction(fd, c, -1);
      return -1;
    }
    if (len > sizeof(c->buf->buf)) {
      buf += len - sizeof(c->buf->buf);
      len = sizeof(c->buf->buf);
    }
    if (max && len > max) {
      buf += len - max;
      len = max;
    }
    memcpy(c->buf->buf, buf, len);
    c->buf->len = len;
  }
  else if (c->buf)
    c->buf->len = 0;
  return 0;
}

static void
hch(int fd, ev_t e, pxyconn_t *c) {
  /* HTTP CONNECT.
   * state:
   *  1 - request sent, want reply
   */
  int l;
  if (pxycerr(fd, e, c)) return;
  if (!c->state) {
    l = sprintf(pxybuf, "CONNECT %s:%d HTTP/1.0\r\n\r\n",
		inet_ntoa(c->dst.addr), c->dst.port);
    if (pxywrite(fd, c, pxybuf, l, 3) == 0) {
      c->state = 1;
      if (pxyrenew(fd, c, 0) == 0)
	pxyev_reqio(fd, EVENT_READ, hch, c);
    }
  }
  else 
    pxyaction(fd, c, 0);
}

static int
hcc(int fd, pxyconn_t *c, char *buf, int l) {
  char *p;
  int r;
  if (c->nread != l)
    return 0;
  if (memmem(buf, l, "HwaRang/", 8))
    pxyinfo(c, 2, "probably HwaRang \"lazy\" proxy");
  if (memcmp(buf, "HTTP/", 5) != 0 ||
      !(p = strchr(buf, ' ')) ||
      p > buf + 8 ||
      p[1] < '1' || p[1] > '9' ||
      p[2] < '0' || p[2] > '9' ||
      p[3] < '0' || p[3] > '9' ||
      p[4] != ' ')
    return 0;
  /* log Proxy-Agent:, Via: etc headers and first line */
  ++p;
  r = (p[0] - '0') * 100 + (p[1] - '0') * 10 + (p[2] - '0');
  if (*p == '2') {
    pxyinfo(c, 2, "HTTP request successeful (%d)", r);
    return 0;
  }
  else {
    /*XX determine whenever this is a permanent or temp error */
    pxyinfo(c, 2, "HTTP request refused or failed (%d)", r);
    if (r == 407 || r == 400 || r == 302)
      pxyaction(fd, c, 2);
    else
      pxyaction(fd, c, 1);
    return 1;
  }
}

static void
s5h(int fd, ev_t e, pxyconn_t *c) {
  /* SOCKS5
   * state:
   *  1 - auth request sent, want reply
   *  2 - connect request sent, want reply
   */
  int l;
  if (pxycerr(fd, e, c)) return;
  if (!c->state) {
    pxybuf[0] = 5; /* socks version */
    pxybuf[1] = 1; /* nmethods */
    pxybuf[2] = 0; /* no auth reqd */
    if (pxywrite(fd, c, pxybuf, 3, 4) == 0) {
      c->state = 1;
      if (pxyev_reqio(fd, EVENT_READ, s5h, c) == 0)
	pxyrenew(fd, c, pxytimeout / 2);
    }
  }
  else if (c->state == 1) {
    if ((l = pxyread(fd, c, pxybuf, 4, 4)) < 2) {
      if (l > 0)
	pxyaction(fd, c, -1);
    }
    else if (pxybuf[1] != 0) {
      pxyinfo(c, 2, "auth rejected (v=%d c=%d%s)",
	      pxybuf[0], pxybuf[1],
	      pxybuf[0] == 0 && pxybuf[1] == 91 ?
	        ", socks4-only server" : "");
      pxyaction(fd, c, 2);
    }
    else if (l > 2) {
      pxyinfo(c, 2, "too much socks5 data v=%d c=%d %d %d",
	      pxybuf[0], pxybuf[1], pxybuf[2], pxybuf[3]);
      pxyaction(fd, c, 2);
    }
    else {
      ipport_t p = htons(c->dst.port);
	pxyinfo(c, 3, "auth accepted");
      pxybuf[0] = 5; /* version */
      pxybuf[1] = 1; /* connect cmd */
      pxybuf[2] = 0; /* reserved */
      pxybuf[3] = 1; /* ipv4 addrtype */
      memcpy(pxybuf+4, &c->dst.addr.s_addr, 4);
      memcpy(pxybuf+8, &p, 2);
      if (pxywrite(fd, c, pxybuf, 10, 4) == 0) {
	c->state = 2;
	pxyrenew(fd, c, 0);
      }
    }
  }
  else if ((l = pxyread(fd, c, pxybuf, 10, 4)) < 2) {
    if (l > 0)
      pxyaction(fd, c, -1);
  }
  else if (pxybuf[1] == 0) {
    pxyinfo(c, 2, "request granted (v=%d c=%d l=%d)",
	    pxybuf[0], pxybuf[1], l);
    pxyaction(fd, c, 0);
  }
  else {
    char *s;
    switch(pxybuf[1]) {
    case 1: s = "general server failure"; l = 1; break;
    case 2: s = "connection not allowed by ruleset"; l = 2; break;
    case 3: s = "network unreachable"; l = 1; break;
    case 4: s = "host unreachable"; l = 1; break;
    case 5: s = "connection refused"; l = 1; break;
    case 6: s = "TTL expired"; l = 1; break;
    case 7: s = "command not supported"; l = 2; break;
    case 8: s = "address type not supported"; l = 2; break;
    default: s = "unexpected response"; l = 1; break;
    }
    pxyinfo(c, 2, "%s (v=%d c=%d)", s, pxybuf[0], pxybuf[1]);
    pxyaction(fd, c, l);
  }
}

static void
s4h(int fd, ev_t e, pxyconn_t *c) {
  /* SOCKS4.
   * state:
   *  1 - connect request sent, want reply
   */
  int l;
  if (pxycerr(fd, e, c)) return;
  if (!c->state) {
    ipport_t p = htons(c->dst.port);
    pxybuf[0] = 4; /* socks version */
    pxybuf[1] = 1; /* request: connect */
    memcpy(pxybuf+2, &p, 2);
    memcpy(pxybuf+4, &c->dst.addr.s_addr, 4);
    pxybuf[8] = 0; /* empty/null username */
    if (pxywrite(fd, c, pxybuf, 9, 4) == 0) {
      c->state = 1;
      if (pxyrenew(fd, c, 0) == 0)
	pxyev_reqio(fd, EVENT_READ, s4h, c);
    }
  }
  else if ((l = pxyread(fd, c, pxybuf, 8, 4)) < 8) {
    if (l > 0)
      pxyaction(fd, c, -1);
  }
  else if (pxybuf[1] == 90) {
    pxyinfo(c, 2, "request granted (v=%d c=%d)", pxybuf[0], pxybuf[1]);
    pxyaction(fd, c, 0);
  }
  else {
    char *s;
    switch(pxybuf[1]) {
    case 91: s = "request rejected or failed"; break;
    case 92: s = "identd required"; break;
    case 93: s = "identd info mismatch"; break;
    case 1:
      if (pxybuf[0] == 5) {
        pxyinfo(c, 2, "request rejected (v=5 c=1, socks5-only server)");
	pxyaction(fd, c, 2);
	return;
      }
    default: s = "unexpected response"; break;
    }
    pxyinfo(c, 2, "%s (v=%d c=%d)", s, pxybuf[0], pxybuf[1]);
    pxyaction(fd, c, 1);
  }
}

static int
wgc(int fd, pxyconn_t *c, char *buf, int l) {
  /* look at Password:/etc strings */
  if (memmem(buf, l, "sername", 7) ||
      memmem(buf, l, "ogin:", 5) ||
      memmem(buf, l, "ssword:", 7)) {
    pxyaction(fd, c, 2);
    return 1;
  }
  else
    return 0;
}

static void
wgh(int fd, ev_t e, pxyconn_t *c) {
  /* WINGATE/telnet.
   * state:
   *  1 - wait for a prompt, interpret Username:/Login: and similar;
   *   we're waiting for a small timeout here, if not found - go to
   *   next state
   *  2 - connection request sent, wait for reply
   */
  int l;
  if (!c->state) {
    if (pxycerr(fd, e, c)) return;
    l = read(fd, pxybuf, 1); /* check if established */
    if (!l) {
      errno = 0;
      pxyaction(fd, c, -1);
    }
    else if (l < 0 && errno != EAGAIN)
      pxyaction(fd, c, -1);
    else {
      c->state = 1;
      if (pxyrenew(fd, c, pxytimeout / 8) == 0)
	pxyev_reqio(fd, EVENT_READ, wgh, c);
    }
  }
  else if (c->state == 1) {
    if (e == EVENT_ERROR && errno == ETIMEDOUT) {
      /*XX try to recognize whenever to use telnet or "plain" wingate */
      l = sprintf(pxybuf, "%s:%d\r\n", inet_ntoa(c->dst.addr), c->dst.port);
      if (c->pxy.port == 23)
	l += sprintf(pxybuf + l, "telnet %s %d\r\n",
		     inet_ntoa(c->dst.addr), c->dst.port);
      if (write(fd, pxybuf, l) < 0)
	pxyaction(fd, c, -1);
      else {
	c->state = 2;
	pxyrenew(fd, c, 0);
      }
    }
    else if (pxycerr(fd, e, c))
      return;
    else if ((l = pxyreadnext(fd, c, 3)) <= 0)
      ;
    else if (wgc(fd, c, pxybuf, l))
      ;
    else
      pxysave(fd, c, pxybuf, l, 0);
  }
  else if (pxycerr(fd, e, c))
    ;
  else
    pxyaction(fd, c, 0);
}

static const ipport_t
htp[] = {
  80,81,1182,3128,4480,6588,7033,8000,8080,8081,8085,8090,8095,8100,8105,8110,
  0 },
spp[] = { 1080,1180,0 },
wgp[] = { 23,1181,0 };

const pxyproto_t pxyprotos[] = {
 { "socks5",  "s5", spp, s5h, NULL },
 { "http",    "hc", htp, hch, hcc },
 { "socks4",  "s4", spp, s4h, NULL },
 { "wingate", "wg", wgp, wgh, wgc },
 {0,0,0,0,0}
};

const pxyproto_t *pxyfindproto(const char *name) {
  const pxyproto_t *p = pxyprotos;
  while(p->name)
    if (strcmp(p->name, name) == 0 || strcmp(p->aname, name) == 0)
      return p;
    else
      ++p;
  return NULL;
}
